@@ -12,14 +12,14 @@ __status__ = "Development"
 
 import random
 
-def classic_mutate_function(island, individual, prob):
+def mutatation_function_classic(individual, prob : float, island=None):
     """
     A Classic mutation function.
 
     Args:
-        island (Island): The Island calling the method.
         individual: Individual object containing a Genome.
         prob: The probability of swapping genes.
+        island (Island): The Island calling the method (optional, default = None).
 
     Returns:
         Individual: The newly mutated individual.
